#! /usr/bin/perl -w

# 
# -------------------------------------------------------------------
#                                   X-BONE
#
#                          http://www.isi.edu/xbone
#                USC Information Sciences Institute (USC/ISI)
#                   Marina del Rey, California 90292, USA
#                          Copyright (c) 1998-2005
# 
# -------------------------------------------------------------------
#
# Copyright (c) 1998-2005 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and
# its documentation in source and binary forms for non-commercial
# purposes and without fee is hereby granted, provided that the above
# copyright notice appear in all copies and that both the copyright
# notice and this permission notice appear in supporting
# documentation, and that any documentation, advertising materials,
# and other materials related to such distribution and use acknowledge
# that the software was developed by the University of Southern
# California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived
# from this software without specific prior written permission.
# 
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT
# THE SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS
# PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so
# noted when applicable.
#
# -------------------------------------------------------------------
#
# Effort partly sponsored by the Defense Advanced Research Projects
# Agency (DARPA) and Air Force Research Laboratory, Air Force Materiel
# Command, USAF, under agreement numbers F30602-98-1-0200 (X-Bone) and
# F30602-01-2-0529 (DynaBone). The views and conclusions contained
# herein are those of the authors and should not be interpreted as
# necessarily representing the official policies or endorsements,
# either expressed or implied, of the Defense Advanced Research
# Projects Agency (DARPA), the Air Force Research Laboratory, or the
# U.S. Government.
#
# This work was partly supported by the NSF STI-XTEND (ANI-0230789)
# and NETFS (ANI-0129689) projects. Any opinions, findings, and
# conclusions or recommendations expressed in this material are those
# of the authors and do not necessarily reflect the views of the
# National Science Foundation.
#
# -------------------------------------------------------------------
#  $RCSfile: prm.pl,v $
#
# $Revision: 1.6 $
#   $Author: pingali $
#     $Date: 2005/03/31 07:03:45 $
#    $State: Exp $
# ----------------------------------------------------------------------------

# This is a protoype of the DynaBone PRM, which multiplexes outerlay
# packets over potentially several innerlays. The current PRM
# prototype has two key limitations:
#
# (1) It assumes knowledge of which innerlay IP addresses belong to
#     the same outerlay host (i.e. a static "BGP/ARP table" in
#     DynaBone project lingo).
#
# (2) It uses a simple, static multiplexing scheme. For details, see
#     the outbound() function below.
#
# When starting the PRM, the outerlay IP address of the local machine
# and the outerlay netmask must be specified. (One PRM manages one
# outerlay; multiple outerlays require multiple PRM processes for
# now.) Additionally, the location of the BARP table file (in Data::Dumper
# format) must be specified on the command line. The innerlay IP
# addresses of the local machine are taken from the BARP table.
#                         _
#                        / |
#                       /  +-- INNERLAY #1 ADDRESS
#                      /   |
#  OUTERLAY IP/MASK --< PRM+-- INNERLAY #2 ADDRESS
#                      \   |
#                       \  +-- INNERLAY #3 ADDRESS
#                        \_|
#
# The PRM operates by creating a "tun" device and configuring it with
# the outerlay IP address and netmask. This "tun" device serves as the
# outgoing interface for applications of the outerlay. It captures all
# outerlay packets and allows the PRM to encapsulate for the
# appropriate innerlay. After re-injection of the encapsulated packet,
# regular kernel outbound processing forwards the packet over the
# appropriate innerlay.
#
# Inbound processing is based on "divert" sockets. During startup, the
# PRM creates one divert rule for each of the specified innerlay
# addresses. All packets arriving at the host matching one of these
# addresses will be diverted to the PRM, which validates and
# decapsulates the innerlay packets. After this, the PRM re-injects
# them into the kernel, and regular inboudn processing continues on
# the decapsulated innerlay (= outerlay) packet.
#
# A state file (in Data::Dumper format) holds information about which
# mux'ing policy the PRM should use, and contains a list of IDs of
# innerlays currently under attack. The PRM will automatically re-read
# this file whenever it changes.
#
# THIS CODE DOES NOT CURRENTLY IMPLEMENT RFC 2003 FULLY, BUT SHOULD!


use strict;
use Getopt::Long;
use IO::File;
use IO::Select;
use IO::Socket::INET;
use FindBin;
use Fcntl;
use NetPacket::IP qw(IP_PROTO_IPIP IP_PROTO_UDP IP_PROTO_TCP IP_MAXPACKET);
use NetPacket::TCP;
use NetPacket::UDP;
use Net::IP qw(:PROC);
use Data::Dumper;
use English;
use Sys::Syslog;
use POSIX;
use File::stat;
use Net::servent;
use Net::protoent;

require "sys/ioctl.ph";
# XXX including these two causes syswarnings, use locally-defined integers
# instead of the constants they contain
# require "net/if_tun.ph";
# require "net/if.ph";


# clear path for taint mode
$ENV{PATH} = "";


# declare some functions
sub sysctl_read ($);


# Constants
#
use constant IFCONFIG  => "/sbin/ifconfig "; # path to ifconfig command
use constant IPFW      => "/sbin/ipfw -q ";  # path to firewall comand
use constant SYSCTL    => "/sbin/sysctl";    # path to sysctl command


# Globals
#
my @innerlays;           # array of innerlay addresses (Net::IP objects)
my @innerlay_ids;        # array of all valid innerlay IDs (shortcut constant)
my @attacked_innerlays;  # list of innerlay IDs currently under attack
my $policy;              # current mux'ing policy
my %barp;                # BGP/ARP table (initialized from $barpfile)


# Command line options
#
my $outerlay;            # IP block (Net::IP object) of the outerlay interface
my $outerlay_addr;       # outerlay IP address of this host (Net::IP object)
my $barpfile;            # file name of the BGP/ARP table
my $statefile;           # file name of the shared state
my $verbose   = 0;       # flag: verbose diagnostics
my $help      = 0;       # print help and exit
my $debug     = 0;       # don't go daemon, and log to STDOUT
my $pidfile   = "/var/run/$FindBin::Script.pid"; # write PID into this file
my $ttl       = sysctl_read "net.inet.ip.ttl";   # TTL for tunnel wrapper


# Purpose:
#       Write error/warning/info message to syslog or stdout, depending
#       on the value of $debug. Syswarn and sysprint return to the
#       caller, sysdie aborts the program.
#
# Arguments:
#       $msg   message to write
#
# Returns:
#       -      (syswarn aborts the program)
#
# Exceptions:
#       -
#
sub syswarn ($;$) {
  my ($msg, $level) = @_;
  $level = "warning" unless defined $level;
  if ($debug) { chomp $msg; print "$msg\n"; }
  else { syslog $level, $msg; }
}
	     
sub sysdie ($) { syswarn shift, "err"; kill TERM => $PID; }
sub sysprint ($) { syswarn shift, "info"; }



# Purpose:
#       Become a daemon. (This code was adapted from Proc::Daemon.)
#
# Arguments:
#       -
#
# Returns:
#       -
#
# Exceptions:
#       -
#
sub daemon () {
  sub _fork {
    if (defined (my $pid = fork)) { return $pid; }
    else { sysdie "Can't fork: $!"; }
  }

  # fork and exit parent
  if (_fork) { exit 0; }

  # detach ourselves from the terminal
  POSIX::setsid or sysdie "Cannot detach from controlling terminal";

  # prevent possibility of acquiring a controlling terminal
  $SIG{'HUP'} = 'IGNORE';
  if (_fork) { exit 0; }
  
  # change working directory
  chdir "/";
  
  # clear file creation mask
  umask 0;
  
  # close open file descriptors
  foreach my $i (0..POSIX::sysconf(&POSIX::_SC_OPEN_MAX)) { POSIX::close $i; }
  
  # reopen stderr, stdout, stdin to /dev/null
  open STDIN,  "+>/dev/null";
  open STDOUT, "+>&STDIN";
  open STDERR, "+>&STDIN";
}



# Purpose:
#       Determine whether two arrays are different (do not contain
#       all identical entries).
#
# Arguments:
#       @array1   first array to compare
#       @array2   second array to compare
#
# Returns:
#       0         if the arrays are identical
#
# Exceptions:
#       -
#
sub different (@@) {
  # Stolen from perlfaq4 question "How do I compute the difference of
  # two arrays? How do I compute the intersection of two arrays?"
  # The original code is actually much more powerful, but has been
  # reduced to a simple difference test.

  my (@a1, @a2) = @_;
  my (@inter, @diff) = ((), ());
  my %c = ();
  
  foreach my $i (@a1, @a2) { $c{$i}++ }
  foreach my $i (keys %c) { push @{ $c{$i} > 1 ? \@inter : \@diff }, $i; }
  return scalar @diff;
}


# Purpose:
#       Translate a string IP_ADDRESS/MASK into a string IP_BLOCK/MASK,
#       where the difference between IP_ADDRESS and IP_BLOCK is that the
#       bits outside the subnet mask are zero in IP_BLOCK. This function
#       is mostly useful when creating Net::IP objects.
#
#       Example: 10.26.8.11/24 -> 10.26.8.0/24
#
# Arguments:
#       $ip   string to translatefirst array to compare
#
# Returns:
#       Translated string.
#
# Exceptions:
#       -
#
sub ip_block ($) {
  my ($ip, $mask) = split "/", shift;
  my $bin_ip = ip_iptobin $ip, 4;
  $bin_ip = substr $bin_ip, 0, $mask;
  $bin_ip .= "0" x (32 - $mask);
  $ip = ip_bintoip $bin_ip,4 ;
  return "$ip/$mask";
}


# Description: 
#       Return value of sysctl variable $var.
#
# Arguments:
#       $var    sysctl variable to check
#
# Returns:
#       value of $var if found
#       undef if $var not found
#
# Exceptions:
#       thrown on failure
#
sub sysctl_read ($) {
  my $var = shift; # variable to read
  my $val;
  
  # get the value by reading from a pipe to sysctl
  my $pipe = SYSCTL ." -n $var 2> /dev/null |";
  open PIPE, $pipe or sysdie "Cannot open pipe $pipe: $!";
  # read the output
  $val = <PIPE>;
  # done with pipe
  close PIPE;
  # post-process
  if(defined $val) { chomp $val; }
  return $val;
}



# Purpose:
#       Print info about command line arguments to stdout and
#       terminate.
#
# Arguments:
#       $msg    error message (omitted or empty -> no error message)
#
# Returns:
#       -
#
# Exceptions:
#       -
#
sub usage ($) {
  my $msg = shift;

  print "ERROR:    $msg\n" if defined $msg and $msg ne "";
  print <<EOB

USAGE:    $FindBin::Script -outerlay <IP/mask> -barp <file> -state <file>


REQUIRED: -outerlay    IP address and netmask of the outerlay
                       (A.B.C.D/E notation)

          -barp        file name of the BGP/ARP table (Data::Dumper format)

          -state       file name of the shared state file (Data::Dumper format)


OPTIONAL: -verbose     verbose mode

          -debug       do not become a daemon, and log to stdout not syslog

          -ttl         TTL to use for the tunnel wrapper
                       (default: $ttl)

          -pidfile     write the PID into this file
                       (default: $pidfile)

Options may be abbreviated with unique prefixes.

EOB
;
  exit;
}



# Purpose:
#       Configure tun to capture outgoing outerlay packets.
#       Must use sysdie/syswarn; this runs after we became a daemon.
#
# Arguments:
#       $outerlay_addr    outerlay IP address (Net::IP object)
#       $outerlay         outerlay block (Net::IP object)
#
# Returns:
#       ($fh, $tundev)    pair of a non-blocking, read/write file handle to
#                         the configured tun device, and its name
#       device
#
# Exceptions:
#       thrown on error
#
sub tun_create ($$) {
  my ($outerlay_addr, $outerlay) = @_;
  
  # Find an available tun device.
  my $i = 0;
  my $cmd = IFCONFIG ."-lu";
  foreach my $if (split / /, `$cmd`) {
    $i++ if $if =~ /tun$i/;
  }
  my $tundev = "tun$i";

  # open the tun device read/write
  my $fh = new IO::File;
  $fh->open("+>/dev/$tundev");
  defined $fh or sysdie "Cannot open /dev/$tundev: $!";
  
  # make it broadcast (default is point-to-point)
  $i = pack "i", 0x2;           # 0x2        is IFF_BROADCAST
  $fh->ioctl(0x8004745e, $i)    # 0x8004745e is TUNSIFMODE
    or sysdie "Cannot ioctl $tundev, TUNSIFMODE, IFF_BROADCAST: $!";
  sysprint "ioctl $tundev, TUNSIFMODE, IFF_BROADCAST\n" if $verbose;
  
  # set the IP address/netmask
  $cmd = join " ", (IFCONFIG, $tundev, $outerlay_addr->ip, 
		    "netmask", $outerlay->mask, "up");
  sysprint $cmd if $verbose;
  system ($cmd) == 0 or sysdie "Cannot execute $cmd: $!";
  
  # make it non-blocking, otherwise later read calls will block
  $fh->autoflush(1);
  fcntl $fh, F_SETFL, O_NONBLOCK 
    or sysdie "Cannot fcntl $tundev, F_SEFTL, O_NONBLOCK: $!";
  sysprint "fcntl $tundev, F_SEFTL, O_NONBLOCK\n" if $verbose;

  # return the file handle (unless exceptions occurred)
  return ($fh, $tundev) unless $@;
}



# Purpose:
#       Destroy tun device (= undo tun_create operation).
#       Must use sysdie/syswarn; this runs after we became a daemon.
#
# Arguments:
#       $fh       file handle of the tun device (return value from
#                 tun_create)
#       $tundev   name of the tun device
#
# Returns:
#       1
#
# Exceptions:
#       -
#
sub tun_destroy ($$) {
  my ($fh, $tundev) = @_;

  # bring interface down and delete its configuration
  foreach my $operation (qw(down delete)) {
    my $cmd = IFCONFIG. "$tundev $operation\n";
    sysprint $cmd if $verbose;
    system ($cmd) == 0 or syswarn "Cannot execute $cmd: $!";
  }

  if (defined $fh) {
    # make interface point-to-point again
    my $i = pack "i", 0x10;     # 0x10       is IFF_POINTOPOINT
    $fh->ioctl(0x8004745e, $i)  # 0x8004745e is TUNSIFMODE
      or syswarn "Cannot ioctl $tundev, TUNSIFMODE, IFF_POINTOPOINT: $!";
    sysprint "ioctl $tundev, TUNSIFMODE, IFF_POINTOPOINT\n" if $verbose;
    
    # close file handle
    $fh->close;
  }
  
  return 1;
}



# Purpose:
#       Set up divert rules to capture all packets arriving over the
#       innerlays.
#       Must use sysdie/syswarn; this runs after we became a daemon.
#
# Arguments:
#       $outerlay      outerlay block (Net::IP object)
#       @innerlays     IP addresses of our local innerlays (Net::IP objects)
#
# Returns:
#       ($fh, @rules)  a pair consisting of a non-blocking, read/write file
#                      handle to the divert socket, and an array of
#                      corresponding ipfw rule numbers used for divert rules
#
# Exceptions:
#       thrown on error
#
sub divert_create ($@) {
  my ($outerlay, @innerlays) = @_;
  my ($port, $fh);

  do {
    # Try random ports until we find an available one. Since the divert
    # port space is usually extremely sparse, this should work fine.
    $port = 1024 + int rand (65535 - 1024);
    
    # Open divert socket.
    sysprint "IO::Socket::INET->new($port, divert, SOCK_RAW)\n" if $verbose;
    $fh = IO::Socket::INET->new(LocalPort => $port,
				Proto     => "divert",
				Type      => SOCK_RAW);
    defined $fh 
      or syswarn "Cannot open divert socket on port $port, retrying: $!";

  } unless (defined $fh);

  # Try to find the two lowest-numbered, available ipfw rule numbers.
  my @rules;
  my %used;
  foreach my $line (IPFW ." show") {
    if ($line =~ /^(\d+)/) { $used{$1} = 1; }
  }
  my $needed = 2;
  for (my $rule = 1; $rule < 65535; $rule++) {
    unless (exists $used{$rule}) {
      push @rules, $rule;
      last if --$needed == 0;
    }
  }

  # We can already create the return value here, so we can mess with
  # the @rules list below.
  my @result = ($fh, @rules);

  # Now that we have an available port and rule numbers, add divert rules.
  my $rule = shift @rules;
  foreach my $innerlay (@innerlays) {
    my $cmd = IPFW. "add $rule divert $port ipencap from any to ". 
      $innerlay->ip ." in\n";
    sysprint $cmd if $verbose;
    system ($cmd) == 0 or sysdie "Cannot execute $cmd: $!";
  }
  
  # Make it non-blocking, otherwise later read calls will block.
  $fh->autoflush(1);
  fcntl $fh, F_SETFL, O_NONBLOCK
    or sysdie "Cannot fcntl divert, F_SEFTL, O_NONBLOCK: $!";

  # Punch a hole into the firewall, to allow all packets over the tun
  # to go through. This needs to occur *after* the divert rules are
  # added, so the decapsed packets pass by this rule and are let
  # through.
  $rule = shift @rules;
  my $block = $outerlay->print;
  my $cmd = IPFW. "add $rule allow ip from $block to $block\n";
  sysprint $cmd if $verbose;
  system ($cmd) == 0 or sysdie "Cannot execute $cmd: $!";

  # Return the divert socket.
  return @result;
}



# Purpose:
#       Destroy the divert rule (undo divert_create operation).
#       Must use sysdie/syswarn; this runs after we became a daemon.
#
# Arguments:
#       $fh     file handle of the divert socket (return value from
#               divert_create)
#
# Returns:
#       1
#
# Exceptions:
#       -
#
sub divert_destroy ($@) {
  my ($fh, @rules) = @_;

  # close file handle if it's open still
  $fh->close if defined $fh;

  # remove divert rules
  foreach my $rule (@rules) {
    my $cmd = IPFW. "delete $rule\n";
    sysprint $cmd if $verbose;
    system ($cmd) == 0 or syswarn "Cannot execute $cmd: $!";
  }
  
  return 1;
}



# Purpose:
#       Return a string that describes the argument packet for debugging.
#
# Arguments:
#       $pkt    packet to describe
#
# Returns:
#       Textual description of the packet.
#
# Exceptions:
#       thrown on error
#
sub describe_pkt ($) {
  
  sub translate_port ($$) {
    my ($port, $proto) = @_;
    return "" unless defined $port;
    my $serv = getservbyport ($port, $proto);
    if (defined $serv) {
      return sprintf ":%d (%s)", $port, $serv->name;
    }
    return sprintf ":%d", $port;
  } 
 
  my ($pkt) = @_;

  # do nothing if not verbose
  return "" unless $verbose;
  
  my ($sport, $dport, $tpkt, $flag_desc);
  
  if ($pkt->{proto} == IP_PROTO_UDP) {
    $tpkt = NetPacket::UDP->decode($pkt->{data});

  } elsif ($pkt->{proto} == IP_PROTO_TCP) {
    $tpkt = NetPacket::TCP->decode($pkt->{data});
    my %flags_desc = (@{[FIN]} => "FIN", @{[SYN]} => "SYN",
		      @{[RST]} => "RST", @{[PSH]} => "PSH", 
		      @{[ACK]} => "ACK", @{[URG]} => "URG",
		      @{[ECE]} => "ECE", @{[CWR]} => "CWR");
    my @flags = ();
    foreach my $f (FIN, SYN, RST, PSH, ACK, URG, ECE, CWR) {
      if ($tpkt->{flags} & $f) { push @flags, $flags_desc{$f}; }
    }
    $flag_desc = "<" . (join "|", @flags) . ">";
  }

  $sport = $tpkt->{src_port};
  $dport = $tpkt->{dest_port};
  $flag_desc = "" unless defined $flag_desc;
  
  my $proto = getprotobynumber $pkt->{proto};
  $proto = $proto->name;
  
  return sprintf "%s%s id %d%s%s %s%s -> %s%s", uc $proto, $flag_desc,
    $pkt->{id}, ($pkt->{flags} == 4 ? " DF" : ""),
    ($pkt->{len} != -1 ? " len ". $pkt->{len} : ""),
    $pkt->{src_ip}, translate_port($sport, $proto), 
    $pkt->{dest_ip}, translate_port($dport, $proto);
}


# Purpose:
#       Pick an innerlay to multiplex over.
#
# Arguments:
#       $pkt     packet to multiplex
#       $policy  mux'ing policy
#
# Returns:
#       Return selected innerlay ID.
#
# Exceptions:
#       thrown on error
#
sub mux ($$) {
  my ($pkt, $policy) = @_;
  my $which = 0;

  # Create a list of currently available innerlays, i.e. all that
  # aren't int he attack list.
  my @ok_innerlays;
  my %attacked = map { $_ => 1 } @attacked_innerlays;
  foreach my $innerlay (@innerlay_ids) {
    push @ok_innerlays, $innerlay unless exists $attacked{$innerlay};
  }

  if ($policy eq "random") {
    # This policy completely randomizes packets over the innerlays.
    $which = int rand scalar @ok_innerlays;

  } elsif ($policy eq "ip-id") {
    # This policy spreads packets into innerlays modulo their IP ID.
    # (On systems where IP IDs are randomized, this is very similar to
    # "random".)
    $which = $pkt->{id} % scalar @ok_innerlays;

  } elsif ($policy eq "round-robin") {
    # This policy does a round-robin on available overlay IDs.

  } else {
    # this should never happen
    sysdie "Illegal policy $policy";
  }

  return $ok_innerlays[$which];
}


# Purpose:
#       Outbound processing - read outgoing outerlay packet from tun
#       device, pick an innerlay to multiplex over, encapsulate (=
#       create an innerlay paacket) and re-inject the result into the
#       kernel.
#
#       Must use sysdie/syswarn; this runs after we became a daemon.
#
# Arguments:
#       $fh     file handle to the tun device from which to read the
#               outgoing packet
#
# Returns:
#       -
#
# Exceptions:
#       thrown on error
#
#
sub outbound ($) {
  my $tun = shift;
  my $raw_pkt = "";

  # Read the outbound packet from the tun device (file handle must be
  # non-blocking, otherwise the read call waits until the passed
  # buffer is full, instead of returning exactly one packet - XXX
  # investigate whether recv works instead of read here)
  sysread $tun, $raw_pkt, IP_MAXPACKET 
    or sysdie "Cannot read packet from tun: $!";
  my $pkt = NetPacket::IP->decode($raw_pkt);

  # Sanity check: RFC 2003 says MUST NOT encapsulate when TTL=0
  if ($pkt->{ttl} == 0) {
    syswarn "Must not encapsulate pkt with ttl=0" if $verbose;
    return;
  }

  # Drop packets destined to the subnet broadcast address of the
  # outerlay.
  if ($pkt->{dest_ip} eq $outerlay->last_ip) {
    syswarn "Dropping broadcast: ". describe_pkt $pkt if $verbose;
    return;
  }

  # If all packets are under attack, drop the packet. Doing this here
  # simplifies the mux function, because it can assume that there will
  # always be at least one innerlay available. (So we can use cheap
  # probabilistic tricks that waste cylces but are easy to implement :-)
  return unless different @innerlay_ids, @attacked_innerlays;

  # Pick an innerlay to forward this packet over.
  my $innerlay = mux $pkt, $policy;

  # Find the innerlay destination for the packet
  my $innerlay_dst = $barp{$pkt->{dest_ip}}{$innerlay};
  unless (defined $innerlay_dst) {
    syswarn "No innerlay $innerlay dst for $pkt->{dest_ip} in BARP"
      if $verbose;
    return;
  }

  # We know which innerlay to encapsulate the packet over, do so now. 
  #
  # THIS CODE MUST DO RFC 2003-CONFORMING IP-IN-IP ENCAPSULATION! (And
  # tries to, but may not cover all borderline cases. Anything
  # different from RFC 2003 should be considered a bug.)
  #
  # The biggest issue we dodge is PMTU discovery-related things, like
  # ICMP relaying/generation. (This is mostly because it's hard to do
  # in userland.) Settig DF on the outer header in all packets seems to
  # help with some PMTU blackholing, but this is still a *major*
  # issue that needs re-thinking.
  #
  # We're reusing a copy of the headers of the original packet, and
  # diddle the fields that are different.
  my $encap_pkt = NetPacket::IP->decode($raw_pkt);
  $encap_pkt->{proto}   = IP_PROTO_IPIP;
  $encap_pkt->{data}    = $raw_pkt;
  $encap_pkt->{flags}   = 4;    # don't fragment (see section 5.1 of RFC 2003)
  $encap_pkt->{src_ip}  = $barp{$outerlay_addr->ip}{$innerlay};
  $encap_pkt->{dest_ip} = $innerlay_dst;
  $encap_pkt->{options} = "";   # RFC 2003 says don't copy IP options
  $encap_pkt->{ttl}     = $ttl; # use specified TTL
  $encap_pkt->{id}      = int rand 65536;
  $encap_pkt->{len}     = -1;   # set by NetPacket::IP, fake so I remember

  # debug info
  if ($verbose) {
    my $pdesc = describe_pkt $pkt;
    my $edesc = describe_pkt $encap_pkt;
    syswarn "Outbound $pdesc inside $edesc over $innerlay.\n"
  }

  # constructed the encapsulated packet, now re-insert it into the
  # kernel (NetPacket::IP::encode automatically computes the checksum
  # and sets the packet length)
  syswrite $tun, $encap_pkt->encode;

}



# Purpose:
#       Inbound processing - read an incoming innerlay packet from the
#       divert socket, validate the encapsulation, decapsulate,
#       validate the contents (= outerlay packet) and re-inject the
#       result back into the kernel.
#
#       Must use sysdie/syswarn; this runs after we became a daemon.
#
# Arguments:
#       $fh     file handle of divert socket from which to read the
#               incoming packet
#
# Returns:
#       -
#
# Exceptions:
#       thrown on error
#
sub inbound ($) {
  my $divert = shift;
  my $raw_pkt = "";

  # a packet is arriving over the innerlay, via a divert rule - read it
  my $tag = recv $divert, $raw_pkt, IP_MAXPACKET, 0 
    or sysdie "Cannot read packet from divert: $!";
  my $pkt = NetPacket::IP->decode($raw_pkt);

  # XXX RFC 2003 may require additional checks here - add them!
  # XXX we should validate whether the incoming packets satisfy the
  # muxing policy in outbound() - we don't do this yet

  # sanity check: make sure pkt is for one of our local innerlay
  # addresses (pkt dest matches one of our local innerlay dests)
  my $found = 0;
  for (my $innerlay = 0; $innerlay < scalar @innerlays; $innerlay++) {
    if ($innerlays[$innerlay]->ip eq $pkt->{dest_ip}) { $found = 1; last; }
  }
  unless ($found) {
    syswarn "Got pkt for non-local innerlay dst $pkt->{dest_ip}" if $verbose;
    goto REINJECT;
  }
  
  # looks like the packet wrapper is OK, decapsulate it
  my $decap_pkt = NetPacket::IP->decode($pkt->{data});

  # sanity check: make sure pkt is for our outerlay address
  unless ($found) {
    syswarn "Got pkt for non-local outerlay dst $decap_pkt->{dest_ip}"
      if $verbose;
    goto REINJECT;
  }

  # sanity check: RFC 2003 says MUST NOT deliver pkt when TTL=0
  if ($decap_pkt->{ttl} == 0) {
    syswarn "Must not deliver pkt with ttl=0" if $verbose;
    return; # and not "goto REINJECT" - this is an invalid packet for us
  }

  # looks like the decaps'ed packet is genuine, re-inject it into the
  # kernel (XXX don't use $divert->send syntax, it reinjects as outbound!)
  send $divert, $pkt->{data}, 0, $tag or sysdie "Cannot re-inject packet: $!";

  # debug info
  if ($verbose) {
    my $pdesc = describe_pkt $pkt;
    my $ddesc = describe_pkt $decap_pkt;
    syswarn "Inbound $ddesc inside $pdesc.\n"
  }

 REINJECT:
  # Re-inject the original raw packet - it might be for someone else!
  send $divert, $raw_pkt, 0, $tag or sysdie "Cannot re-inject raw packet: $!";
}




# Purpose:
#       Re-load the state if it has changed on the disk.
# 
#       Must use sysdie/syswarn; this runs after we became a daemon.
#
# Arguments:
#       $outerlay_addr    host's IP address in the outerlay (Net::IP object)
#
# Returns:
#       -
#
# Exceptions:
#       thrown on error
#

# Modification timestamp of the state file.
my $state_file_last_modified = 0;

sub reload_state ($) {
  my ($outerlay_addr) = @_;
  my %state;
  my $st = stat $statefile;

  if ($st->mtime > $state_file_last_modified) {
    eval {
      # We need to re-read the state file.
      open FH, $statefile or die "Cannot open $statefile: $!";
      my $statetext = "";
      while (<FH>) { chomp; $statetext .= $_; }
      close FH or die "Cannot close $statefile: $!";
      
      # untaint state content and evaluate it (XXX security hole!)
      if ($statetext =~ /(.*)/) { $statetext = $1; }
      else { die "Invalid state contents"; }
      eval $statetext or die "Error evaluting state: $@";
      
      # sanity check: attack list must contain only known innerlay IDs
      foreach my $id ( @{ $state{attacked_innerlays} }) {
	die "State contains unknown attacked innerlay ID $id"
	  unless exists $barp{$outerlay_addr->ip}{$id};
      }
      
      # sanity check: valid policy
      die "State contains unknown policy $state{policy}"
	unless $state{policy} =~ /^(random|round-robin|ip-id)$/;
    };

    if ($@) {
      # issue with state, don't replace the current one, just warn
      syswarn "Not updating from invalid state: $@";

    } else {
      # changed state is OK, replace current one
      $policy = $state{policy};
      @attacked_innerlays = @{ $state{attacked_innerlays} };

      $state_file_last_modified = $st->mtime;

      syswarn "State reloaded\n" if $verbose;
    }
  }
}
  
  
  
  
# Main body. First we handle our command line arguments.
if (GetOptions("barp=s"     => \$barpfile,
               "state=s"    => \$statefile, 
	       "outerlay=s" => \$outerlay,
	       "pidfile=s"  => \$pidfile,
	       "verbose!"   => \$verbose,
	       "debug!"     => \$debug,
	       "help"       => \$help) == 0) { usage ""; }

# did the user specify -help? if so, print and exit, no matter what
# else
usage "" if $help;

# check pidfile argument
if ($pidfile =~ /^(.*)$/) { $pidfile = $1; } # XXX how to best untaint a path?
else { usage "Invalid PID file $pidfile"; }
usage "PID file $pidfile already exists" if -f $pidfile;
	     
# need both outerlay and at least one innerlay
usage "Specify one outerlay, one BARP table, and one state file."
  unless defined $barpfile and -f $barpfile 
     and defined $statefile and -f $statefile and defined $outerlay;

# outerlay needs to be in XXX.XXX.XXX.XXX/XX notation (and de-taint)
if ($outerlay =~ /^((\d{1,3}\.){3}\d{1,3}\/\d{1,2})$/) { $outerlay = $1; }
else { usage "Invalid outerlay $outerlay"; } 
($outerlay_addr) = split "/", $outerlay;
$outerlay_addr = new Net::IP $outerlay_addr or die Net::IP::Error;
$outerlay = new Net::IP ip_block $outerlay or die Net::IP::Error;

# Make sure forwarding is enabled, otherwise the PRM will fail.
sysdie "IP forwarding not enabled" 
  unless sysctl_read "net.inet.ip.forwarding" == 1;

# read in the BARP table
open FH, $barpfile or die "Cannot open $barpfile: $!";
my $barptext = "";
while (<FH>) { chomp; $barptext .= $_; }
close FH or die "Cannot close $barpfile: $!";

# untaint BARP file contents and evaluate the contents
# XXX security hole!
if ($barptext =~ /(.*)/) { $barptext = $1; }
else { usage "Invalid BARP table contents"; }
eval $barptext or die "Error evaluting BARP table: $@";

# sanity check: BARP table must contain local outerlay IP address
usage "BARP table does not contain outerlay address $outerlay_addr->ip"
  unless exists $barp{$outerlay_addr->ip};

# Initialize the global that holds all innerlay IDs (used as shortcut constant)
@innerlay_ids = keys %{ $barp{$outerlay_addr->ip} };

# sanity check: innerlay IDs in all BARP entries must be identical
foreach my $o (keys %barp) {
  usage "some BARP entries have different innerlay IDs"
    if different @innerlay_ids, keys %{ $barp{$o} }
}

# sanity check: all innerlay IPs must be different
my %ips;
foreach my $o (keys %barp) {
  foreach my $i (@innerlay_ids) {
    usage "Non-unique innerlay IP $barp{$o}{$i} in BARP"
      if exists $ips{$barp{$o}{$i}};
    $ips{$barp{$o}{$i}} = 1;
  }
}

# Set innerlays array from BARP (easier than re-writing everyhing to
# use BARP table directly, for now). Sort the innerlay IP addresses
# according to the alphabetical order of their corresponding innerlay
# IDs, so that all nodes participating in an overlay agree on the same
# order of innerlays, this is used when multiplexing.
foreach my $i (@innerlay_ids) {
  push @innerlays, new Net::IP $barp{$outerlay_addr->ip}{$i};
}

# open syslog
Sys::Syslog::setlogsock "unix" or die "Cannot set AF_UNIX syslog: $!";
openlog $FindBin::Script, "cons,pid", "user" or die "Cannot open syslog: $!";

# become a daemon (unless debug is set) - FROM HERE ON OUT, USE SYSWARN/SYSDIE
daemon unless $debug;

# write PID to pidfile
open FH, ">$pidfile" or sysdie "Cannot open $pidfile: $!";
print FH $PID;
close FH or sysdie "Cannot close $pidfile: $!";

# Done handling command line options. Set up tun device and divert
# rules now.
my ($tun, $tundev, $divert, @rules);

# Try configuration...
eval { 
  ($tun, $tundev) = tun_create $outerlay_addr, $outerlay;
  ($divert, @rules) = divert_create $outerlay, @innerlays;
};
# ...and do this block (PRM main loop) if all went well above,
# otherwise skip it, and tear everything down before exiting...
unless ($@) {
  # install a signal handler for a forceful exit
  my $done = 0;
  $SIG{INT} = $SIG{TERM} = sub { $done = 1 };

  # Select loop: listen on tun device and divert socket and handle packets
  my $select = new IO::Select($tun, $divert);
  while (not $done and my @ready = $select->can_read) {

    # Before we go off and handle the packet, first check whether the
    # state file has been updated, and if so, re-read it.
    reload_state $outerlay_addr;

    foreach my $fh (@ready) {
      if ($fh == $tun) {
	# outbound outerlay packet arrived on tun device 
	# -> outbound processing (exceptions handled below)
	eval { outbound $fh; }
      } elsif ($fh == $divert) {
	# inbound innerlay packet arrived
	# -> inbound processing (exceptions handled below)
	eval { inbound $fh; };
      } else {
	# this should never happen
	eval { die "Unknown handle"; }
      }
      # any exceptions? then we're done
      syswarn $@ and last if ($@);
    }
  }
}
	     
# clean up, we're done!
divert_destroy $divert, @rules;
tun_destroy $tun, $tundev;
unlink $pidfile or syswarn "Cannot unlink $pidfile: $!";
closelog;
exit;
